# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
import logging
# non-stdlib imports
import azure.storage.blob
# local imports
import blobxfer.retry

# create logger
logger = logging.getLogger(__name__)


def create_client(storage_account, timeout):
    # type: (blobxfer.operations.azure.StorageAccount,
    #        tuple) -> BlockBlobService
    """Create block blob client
    :param blobxfer.operations.azure.StorageAccount storage_account:
        storage account
    :param tuple timeout: timeout tuple
    :rtype: azure.storage.blob.BlockBlobService
    :return: block blob service client
    """
    if storage_account.is_sas:
        client = azure.storage.blob.BlockBlobService(
            account_name=storage_account.name,
            sas_token=storage_account.key,
            endpoint_suffix=storage_account.endpoint,
            request_session=storage_account.session,
            socket_timeout=timeout)
    else:
        client = azure.storage.blob.BlockBlobService(
            account_name=storage_account.name,
            account_key=storage_account.key,
            endpoint_suffix=storage_account.endpoint,
            request_session=storage_account.session,
            socket_timeout=timeout)
    # set retry policy
    client.retry = blobxfer.retry.ExponentialRetryWithMaxWait().retry
    return client


def create_blob(ase, data, md5, metadata, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, bytes, str, dict,
    #        int) -> None
    """Create one shot block blob
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param bytes data: blob data
    :param str md5: md5 as base64
    :param dict metadata: metadata kv pairs
    :param int timeout: timeout
    """
    ase.client._put_blob(
        container_name=ase.container,
        blob_name=ase.name,
        blob=data,
        content_settings=azure.storage.blob.models.ContentSettings(
            content_type=blobxfer.util.get_mime_type(ase.name),
            content_md5=md5,
        ),
        metadata=metadata,
        validate_content=False,  # integrity is enforced with HTTPS
        timeout=timeout)  # noqa


def _format_block_id(chunk_num):
    # type: (int) -> str
    """Create a block id given a block (chunk) number
    :param int chunk_num: chunk number
    :rtype: str
    :return: block id
    """
    return '{0:08d}'.format(chunk_num)


def put_block(ase, offsets, data, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity,
    #        blobxfer.models.upload.Offsets, bytes, int) -> None
    """Puts a block into remote blob
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param blobxfer.models.upload.Offsets offsets: upload offsets
    :param bytes data: data
    :param int timeout: timeout
    """
    ase.client.put_block(
        container_name=ase.container,
        blob_name=ase.name,
        block=data,
        block_id=_format_block_id(offsets.chunk_num),
        validate_content=False,  # integrity is enforced with HTTPS
        timeout=timeout)  # noqa


def put_block_list(ase, last_block_num, md5, metadata, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, bytes, str, dict,
    #        int) -> None
    """Create block blob from blocks
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param int last_block_num: last block number (chunk_num)
    :param str md5: md5 as base64
    :param dict metadata: metadata kv pairs
    :param int timeout: timeout
    """
    # construct block list
    block_list = [
        azure.storage.blob.BlobBlock(id=_format_block_id(x))
        for x in range(0, last_block_num + 1)
    ]
    ase.client.put_block_list(
        container_name=ase.container,
        blob_name=ase.name,
        block_list=block_list,
        content_settings=azure.storage.blob.models.ContentSettings(
            content_type=blobxfer.util.get_mime_type(ase.name),
            content_md5=md5,
        ),
        metadata=metadata,
        validate_content=False,  # integrity is enforced with HTTPS
        timeout=timeout)


def get_committed_block_list(ase, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, int) -> list
    """Get committed block list
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param int timeout: timeout
    :rtype: list
    :return: list of committed blocks
    """
    if blobxfer.util.blob_is_snapshot(ase.name):
        blob_name, snapshot = blobxfer.util.parse_blob_snapshot_parameter(
            ase.name)
    else:
        blob_name = ase.name
        snapshot = None
    return ase.client.get_block_list(
        container_name=ase.container,
        blob_name=blob_name,
        snapshot=snapshot,
        block_list_type=azure.storage.blob.BlockListType.Committed,
        timeout=timeout).committed_blocks



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function  # noqa
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
# non-stdlib imports
# local imports

from .operations.azure.blob.append import (  # noqa
    create_client as create_append_blob_client
)
from .operations.azure.blob.block import (  # noqa
    create_client as create_block_blob_client
)
from .operations.azure.blob.page import (  # noqa
    create_client as create_page_blob_client
)
from .operations.azure.file import (  # noqa
    create_client as create_file_client
)

from .operations.download import (  # noqa
    Downloader
)
from .operations.synccopy import (  # noqa
    SyncCopy
)
from .operations.upload import (  # noqa
    Uploader
)



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import collections
import json
import logging
# non-stdlib imports
# local imports
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
JSON_KEY_BLOBXFER_METADATA = 'blobxfer_metadata'
# file attributes
_JSON_KEY_FILE_ATTRIBUTES = 'FileAttributes'
_JSON_KEY_FILE_ATTRIBUTES_POSIX = 'POSIX'
_JSON_KEY_FILE_ATTRIBUTES_WINDOWS = 'Windows'
_JSON_KEY_FILE_ATTRIBUTES_MODE = 'mode'
_JSON_KEY_FILE_ATTRIBUTES_UID = 'uid'
_JSON_KEY_FILE_ATTRIBUTES_GID = 'gid'
# vectored io
_JSON_KEY_VECTORED_IO = 'VectoredIO'
_JSON_KEY_VECTORED_IO_MODE = 'Mode'
_JSON_KEY_VECTORED_IO_STRIPE = 'Stripe'
_JSON_KEY_VECTORED_IO_STRIPE_TOTAL_SIZE = 'TotalSize'
_JSON_KEY_VECTORED_IO_STRIPE_OFFSET_START = 'OffsetStart'
_JSON_KEY_VECTORED_IO_STRIPE_TOTAL_SLICES = 'TotalSlices'
_JSON_KEY_VECTORED_IO_STRIPE_SLICE_ID = 'SliceId'
_JSON_KEY_VECTORED_IO_STRIPE_NEXT = 'Next'
# named tuples
PosixFileAttr = collections.namedtuple(
    'PosixFileAttr', [
        'gid',
        'mode',
        'uid',
    ]
)
WindowsFileAttr = collections.namedtuple(
    'WindowsFileAttr', [
    ]
)
VectoredStripe = collections.namedtuple(
    'VectoredStripe', [
        'next',
        'offset_start',
        'slice_id',
        'total_size',
        'total_slices',
    ]
)
VectoredNextEntry = collections.namedtuple(
    'VectoredNextEntry', [
        'storage_account_name',
        'endpoint',
        'container',
        'name',
    ]
)


def get_md5_from_metadata(ase):
    # type: (blobxfer.models.azure.StorageEntity) -> str
    """Get MD5 from properties or metadata
    :param blobxfer.models.azure.StorageEntity ase: Azure Storage Entity
    :rtype: str or None
    :return: md5
    """
    # if encryption metadata is present, check for pre-encryption
    # md5 in blobxfer extensions
    md5 = None
    if ase.is_encrypted:
        md5 = ase.encryption_metadata.blobxfer_extensions.\
            pre_encrypted_content_md5
    if blobxfer.util.is_none_or_empty(md5):
        md5 = ase.md5
    return md5


def generate_fileattr_metadata(local_path, metadata):
    # type: (blobxfer.models.upload.LocalPath, dict) -> dict
    """Generate file attribute metadata dict
    :param blobxfer.models.upload.LocalPath local_path: local path
    :param dict metadata: existing metadata dict
    :rtype: dict
    :return: merged metadata dictionary
    """
    if blobxfer.util.on_windows():
        logger.warning(
            'file attributes store/restore on Windows is not supported yet')
        return None
    else:
        md = {
            _JSON_KEY_FILE_ATTRIBUTES: {
                _JSON_KEY_FILE_ATTRIBUTES_POSIX: {
                    _JSON_KEY_FILE_ATTRIBUTES_MODE: local_path.mode,
                    _JSON_KEY_FILE_ATTRIBUTES_UID: local_path.uid,
                    _JSON_KEY_FILE_ATTRIBUTES_GID: local_path.gid,
                }
            }
        }
        return blobxfer.util.merge_dict(metadata, md)


def fileattr_from_metadata(md):
    # type: (dict) -> collections.namedtuple
    """Convert fileattr metadata in json metadata
    :param dict md: metadata dictionary
    :rtype: PosixFileAttr or WindowsFileAttr or None
    :return: fileattr metadata
    """
    try:
        mdattr = json.loads(
            md[JSON_KEY_BLOBXFER_METADATA])[_JSON_KEY_FILE_ATTRIBUTES]
    except (KeyError, TypeError):
        return None
    else:
        if blobxfer.util.on_windows():
            logger.warning(
                'file attributes store/restore on Windows is not supported '
                'yet')
            fileattr = None
        else:
            try:
                fileattr = PosixFileAttr(
                    mode=mdattr[_JSON_KEY_FILE_ATTRIBUTES_POSIX][
                        _JSON_KEY_FILE_ATTRIBUTES_MODE],
                    uid=mdattr[_JSON_KEY_FILE_ATTRIBUTES_POSIX][
                        _JSON_KEY_FILE_ATTRIBUTES_UID],
                    gid=mdattr[_JSON_KEY_FILE_ATTRIBUTES_POSIX][
                        _JSON_KEY_FILE_ATTRIBUTES_GID],
                )
            except KeyError:
                fileattr = None
        return fileattr


def create_vectored_io_next_entry(ase):
    # type: (blobxfer.models.azure.StorageEntity) -> str
    """Create Vectored IO next entry id
    :param blobxfer.models.azure.StorageEntity ase: Azure Storage Entity
    :rtype: str
    :return: vectored io next entry
    """
    return ';'.join(
        (ase.client.primary_endpoint, ase.container, ase.name)
    )


def explode_vectored_io_next_entry(entry):
    # type: (str, int) -> str
    """Explode next vectored io entry
    :param str entry: next entry
    :rtype: VectoredNextEntry
    :return: vectored next entry
    """
    tmp = entry.split(';')
    _sa = tmp[0].split('.')
    return VectoredNextEntry(
        storage_account_name=_sa[0],
        endpoint='.'.join(_sa[2:]),
        container=tmp[1],
        name=tmp[2],
    )


def remove_vectored_io_slice_suffix_from_name(name, slice):
    # type: (str, int) -> str
    """Remove vectored io (stripe) slice suffix from a given name
    :param str name: entity name
    :param int slice: slice num
    :rtype: str
    :return: name without suffix
    """
    suffix = '.bxslice-{}'.format(slice)
    if name.endswith(suffix):
        return name[:-len(suffix)]
    else:
        return name


def generate_vectored_io_stripe_metadata(local_path, metadata):
    # type: (blobxfer.models.upload.LocalPath, dict) -> dict
    """Generate vectored io stripe metadata dict
    :param blobxfer.models.upload.LocalPath local_path: local path
    :param dict metadata: existing metadata dict
    :rtype: dict
    :return: merged metadata dictionary
    """
    md = {
        _JSON_KEY_VECTORED_IO: {
            _JSON_KEY_VECTORED_IO_MODE: _JSON_KEY_VECTORED_IO_STRIPE,
            _JSON_KEY_VECTORED_IO_STRIPE: {
                _JSON_KEY_VECTORED_IO_STRIPE_TOTAL_SIZE: local_path.total_size,
                _JSON_KEY_VECTORED_IO_STRIPE_OFFSET_START:
                local_path.view.fd_start,
                _JSON_KEY_VECTORED_IO_STRIPE_TOTAL_SLICES:
                local_path.view.total_slices,
                _JSON_KEY_VECTORED_IO_STRIPE_SLICE_ID:
                local_path.view.slice_num,
                _JSON_KEY_VECTORED_IO_STRIPE_NEXT: local_path.view.next,
            }
        }
    }
    return blobxfer.util.merge_dict(metadata, md)


def vectored_io_from_metadata(md):
    # type: (dict) -> collections.namedtuple
    """Convert vectored io metadata in json metadata
    :param dict md: metadata dictionary
    :rtype: VectoredStripe or None
    :return: vectored io metadata
    """
    try:
        mdattr = json.loads(
            md[JSON_KEY_BLOBXFER_METADATA])[_JSON_KEY_VECTORED_IO]
    except (KeyError, TypeError):
        pass
    else:
        if mdattr[_JSON_KEY_VECTORED_IO_MODE] == _JSON_KEY_VECTORED_IO_STRIPE:
            mdstripe = mdattr[_JSON_KEY_VECTORED_IO_STRIPE]
            try:
                nextptr = explode_vectored_io_next_entry(
                    mdstripe[_JSON_KEY_VECTORED_IO_STRIPE_NEXT])
            except (KeyError, AttributeError):
                nextptr = None
            vio = VectoredStripe(
                total_size=mdstripe[_JSON_KEY_VECTORED_IO_STRIPE_TOTAL_SIZE],
                offset_start=mdstripe[
                    _JSON_KEY_VECTORED_IO_STRIPE_OFFSET_START],
                total_slices=mdstripe[
                    _JSON_KEY_VECTORED_IO_STRIPE_TOTAL_SLICES],
                slice_id=mdstripe[_JSON_KEY_VECTORED_IO_STRIPE_SLICE_ID],
                next=nextptr,
            )
            return vio
        else:
            raise RuntimeError('Cannot handle Vectored IO mode: {}'.format(
                mdattr[_JSON_KEY_VECTORED_IO_MODE]))
    return None



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import collections
import logging
import multiprocessing
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
# non-stdlib imports
# local imports
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
_DEFAULT_REQUESTS_TIMEOUT = (3.1, 12.1)

# named tuples
VectoredIo = collections.namedtuple(
    'VectoredIoOptions', [
        'stripe_chunk_size_bytes',
        'distribution_mode',
    ]
)
SkipOn = collections.namedtuple(
    'SkipOn', [
        'filesize_match',
        'lmt_ge',
        'md5_match',
    ]
)
FileProperties = collections.namedtuple(
    'FileProperties', [
        'attributes',
        'md5',
    ]
)
Upload = collections.namedtuple(
    'Upload', [
        'chunk_size_bytes',
        'delete_extraneous_destination',
        'mode',
        'one_shot_bytes',
        'overwrite',
        'recursive',
        'rename',
        'rsa_public_key',
        'store_file_properties',
        'strip_components',
        'vectored_io',
    ]
)
Download = collections.namedtuple(
    'Download', [
        'check_file_md5',
        'chunk_size_bytes',
        'delete_extraneous_destination',
        'mode',
        'overwrite',
        'recursive',
        'rename',
        'restore_file_attributes',
        'rsa_private_key',
    ]
)
SyncCopy = collections.namedtuple(
    'SyncCopy', [
        'delete_extraneous_destination',
        'dest_mode',
        'mode',
        'overwrite',
        'recursive',
    ]
)


class Timeout(object):
    """Timeout Options"""
    def __init__(self, connect, read):
        """Ctor for Timeout options
        :param Timeout self: this
        :param float connect: connect timeout
        :param float read: read timeout
        """
        if connect is None or connect <= 0:
            self._connect = _DEFAULT_REQUESTS_TIMEOUT[0]
        else:
            self._connect = connect
        if read is None or read <= 0:
            self._read = _DEFAULT_REQUESTS_TIMEOUT[1]
        else:
            self._read = read

    @property
    def connect(self):
        """Connect timeout
        :rtype: float
        :return: connect timeout
        """
        return self._connect

    @property
    def read(self):
        """Read timeout
        :rtype: float
        :return: read timeout
        """
        return self._read

    @property
    def timeout(self):
        """Timeout property in requests format
        :rtype: tuple
        :return: (connect, read) timeout tuple
        """
        return (self._connect, self._read)


class Concurrency(object):
    """Concurrency Options"""
    def __init__(
            self, crypto_processes, md5_processes, disk_threads,
            transfer_threads, action=None):
        """Ctor for Concurrency Options
        :param Concurrency self: this
        :param int crypto_processes: number of crypto procs
        :param int md5_processes: number of md5 procs
        :param int disk_threads: number of disk threads
        :param int transfer_threads: number of transfer threads
        :param int action: action hint (1=Download, 2=Upload, 3=SyncCopy)
        """
        self.crypto_processes = crypto_processes
        self.md5_processes = md5_processes
        self.disk_threads = disk_threads
        self.transfer_threads = transfer_threads
        # allow crypto processes to be zero (which will inline crypto
        # routines with main process)
        if self.crypto_processes is None or self.crypto_processes < 1:
            self.crypto_processes = 0
        if self.md5_processes is None or self.md5_processes < 1:
            self.md5_processes = multiprocessing.cpu_count() >> 1
        if self.md5_processes < 1:
            self.md5_processes = 1
        auto_disk = False
        if self.disk_threads is None or self.disk_threads < 1:
            self.disk_threads = multiprocessing.cpu_count() << 1
            # cap maximum number of disk threads from cpu count to 64
            if self.disk_threads > 64:
                self.disk_threads = 64
            # for download action, cap disk threads to lower value
            if action == 1 and self.disk_threads > 16:
                self.disk_threads = 16
            auto_disk = True
        # for synccopy action, set all non-transfer counts to zero
        if action == 3:
            auto_disk = False
            self.md5_processes = 0
            self.crypto_processes = 0
            self.disk_threads = 0
        if self.transfer_threads is None or self.transfer_threads < 1:
            if auto_disk:
                self.transfer_threads = self.disk_threads << 1
            else:
                self.transfer_threads = multiprocessing.cpu_count() << 2
            # cap maximum number of threads from cpu count to 96
            if self.transfer_threads > 96:
                self.transfer_threads = 96


class General(object):
    """General Options"""
    def __init__(
            self, concurrency, log_file=None, progress_bar=True,
            resume_file=None, timeout=None, verbose=False):
        """Ctor for General Options
        :param General self: this
        :param Concurrency concurrency: concurrency options
        :param bool progress_bar: progress bar
        :param str log_file: log file
        :param str resume_file: resume file
        :param Timeout timeout: timeout options
        :param bool verbose: verbose output
        """
        if concurrency is None:
            raise ValueError('concurrency option is unspecified')
        self.concurrency = concurrency
        self.log_file = log_file
        self.progress_bar = progress_bar
        if blobxfer.util.is_not_empty(resume_file):
            self.resume_file = pathlib.Path(resume_file)
        else:
            self.resume_file = None
        self.timeout = timeout
        self.verbose = verbose



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
import logging
try:
    import queue
except ImportError:  # noqa
    import Queue as queue
# non-stdlib imports
# local imports
import blobxfer.models.azure
import blobxfer.models.offload
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
_EMPTY_MAX_PAGE_SIZE_MD5 = 'tc+p1sj+vWGPkawoQ9UKHA=='
_MAX_PAGE_SIZE_BYTES = 4194304


def compute_md5_for_file_asbase64(
        filename, pagealign=False, start=None, end=None, blocksize=65536):
    # type: (str, bool, int, int, int) -> str
    """Compute MD5 hash for file and encode as Base64
    :param str filename: file to compute MD5 for
    :param bool pagealign: page align data
    :param int start: file start offset
    :param int end: file end offset
    :param int blocksize: block size
    :rtype: str
    :return: MD5 for file encoded as Base64
    """
    hasher = blobxfer.util.new_md5_hasher()
    with open(filename, 'rb') as filedesc:
        if start is not None:
            filedesc.seek(start)
            curr = start
        else:
            curr = 0
        while True:
            if end is not None and curr + blocksize > end:
                blocksize = end - curr
            if blocksize == 0:
                break
            buf = filedesc.read(blocksize)
            if not buf:
                break
            buflen = len(buf)
            if pagealign and buflen < blocksize:
                aligned = blobxfer.util.page_align_content_length(buflen)
                if aligned != buflen:
                    buf = buf.ljust(aligned, b'\0')
            hasher.update(buf)
            curr += blocksize
        return blobxfer.util.base64_encode_as_string(hasher.digest())


def compute_md5_for_data_asbase64(data):
    # type: (obj) -> str
    """Compute MD5 hash for bits and encode as Base64
    :param any data: data to compute MD5 for
    :rtype: str
    :return: MD5 for data
    """
    hasher = blobxfer.util.new_md5_hasher()
    hasher.update(data)
    return blobxfer.util.base64_encode_as_string(hasher.digest())


def check_data_is_empty(data):
    # type: (bytes) -> bool
    """Check if data is empty via MD5
    :param bytes data: data to check
    :rtype: bool
    :return: if data is empty
    """
    contentmd5 = compute_md5_for_data_asbase64(data)
    datalen = len(data)
    if datalen == _MAX_PAGE_SIZE_BYTES:
        if contentmd5 == _EMPTY_MAX_PAGE_SIZE_MD5:
            return True
    else:
        data_chk = b'\0' * datalen
        if compute_md5_for_data_asbase64(data_chk) == contentmd5:
            return True
    return False


class LocalFileMd5Offload(blobxfer.models.offload._MultiprocessOffload):
    """LocalFileMd5Offload"""
    def __init__(self, num_workers):
        # type: (LocalFileMd5Offload, int) -> None
        """Ctor for Local File Md5 Offload
        :param LocalFileMd5Offload self: this
        :param int num_workers: number of worker processes
        """
        super(LocalFileMd5Offload, self).__init__(
            self._worker_process, num_workers, 'MD5')

    def _worker_process(self):
        # type: (LocalFileMd5Offload) -> None
        """Compute MD5 for local file
        :param LocalFileMd5Offload self: this
        """
        while not self.terminated:
            try:
                key, lpath, fpath, remote_md5, pagealign, lpview = \
                    self._task_queue.get(True, 0.1)
            except queue.Empty:
                continue
            if lpview is None:
                start = None
                end = None
                size = None
            else:
                start = lpview.fd_start
                end = lpview.fd_end
                size = end - start
            md5 = blobxfer.operations.md5.compute_md5_for_file_asbase64(
                fpath, pagealign, start, end)
            logger.debug('pre-transfer MD5 check: {} <L..R> {} {}'.format(
                md5, remote_md5, fpath))
            self._done_cv.acquire()
            self._done_queue.put((key, lpath, size, md5 == remote_md5))
            self._done_cv.notify()
            self._done_cv.release()

    def add_localfile_for_md5_check(
            self, key, lpath, fpath, remote_md5, mode, lpview):
        # type: (LocalFileMd5Offload, str, str, str, str,
        #        blobxfer.models.azure.StorageModes, object) -> None
        """Add a local file to MD5 check queue
        :param LocalFileMd5Offload self: this
        :param str key: md5 map key
        :param str lpath: "local" path for descriptor
        :param str fpath: "final" path for/where file
        :param str remote_md5: remote MD5 to compare against
        :param blobxfer.models.azure.StorageModes mode: mode
        :param object lpview: local path view
        """
        if blobxfer.util.is_none_or_empty(remote_md5):
            raise ValueError('comparison MD5 is empty for file {}'.format(
                lpath))
        if mode == blobxfer.models.azure.StorageModes.Page:
            pagealign = True
        else:
            pagealign = False
        self._task_queue.put(
            (key, lpath, fpath, remote_md5, pagealign, lpview)
        )



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import collections
import enum
import json
import logging
import math
import os
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
import threading
# non-stdlib imports
import bitstring
# local imports
import blobxfer.models
import blobxfer.models.azure
import blobxfer.models.crypto
import blobxfer.models.metadata
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
_MAX_BLOCK_BLOB_ONESHOT_BYTES = 268435456
_MAX_BLOCK_BLOB_CHUNKSIZE_BYTES = 104857600
_MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES = 4194304
_MAX_NUM_CHUNKS = 50000
_MAX_PAGE_BLOB_SIZE = 8796093022208
_DEFAULT_AUTO_CHUNKSIZE_BYTES = 16777216
_MD5_CACHE_RESUME_ENTRIES_GC_THRESHOLD = 25


# named tuples
Offsets = collections.namedtuple(
    'Offsets', [
        'chunk_num',
        'num_bytes',
        'range_end',
        'range_start',
        'pad',
    ]
)
LocalPathView = collections.namedtuple(
    'LocalPathView', [
        'fd_end',
        'fd_start',
        'mode',
        'next',
        'slice_num',
        'total_slices',
    ]
)


class VectoredIoDistributionMode(enum.Enum):
    Disabled = 'disabled'
    Stripe = 'stripe'
    Replica = 'replica'

    def __str__(self):
        return self.value


class LocalPath(object):
    """Local Path"""

    def __init__(self, parent_path, relative_path, use_stdin=False, view=None):
        # type: (LocalPath, pathlib.Path, pathlib.Path, bool,
        #        LocalPathView) -> None
        """Ctor for LocalPath
        :param LocalPath self: this
        :param pathlib.Path parent_path: parent path
        :param pathlib.Path relative_path: relative path
        :param bool use_stdin: use stdin
        :param LocalPathView view: local path view
        """
        self.parent_path = parent_path
        self.relative_path = relative_path
        self.use_stdin = use_stdin
        # populate properties
        if self.use_stdin:
            # create dummy stat object
            self._stat = lambda: None
            self._stat.st_size = 0
            self._stat.st_mtime = 0
            self._stat.st_mode = 0
            self._stat.st_uid = 0
            self._stat.st_gid = 0
        else:
            self._stat = self.absolute_path.stat()
        if view is None:
            self.view = LocalPathView(
                fd_start=0,
                fd_end=self._stat.st_size,
                slice_num=0,
                mode=VectoredIoDistributionMode.Disabled,
                total_slices=1,
                next=None,
            )
        else:
            self.view = view
        self._size = self.view.fd_end - self.view.fd_start

    @property
    def absolute_path(self):
        # type: (LocalPath) -> pathlib.Path
        """Absolute path
        :param LocalPath self: this
        :rtype: pathlib.Path
        :return: absolute path
        """
        return self.parent_path / self.relative_path

    @property
    def size(self):
        # type: (LocalPath) -> int
        """Size of view
        :param LocalPath self: this
        :rtype: int
        :return: size of view portion of the file
        """
        return self._size

    @property
    def total_size(self):
        # type: (LocalPath) -> int
        """Total Size of file
        :param LocalPath self: this
        :rtype: int
        :return: total size of file (non-view)
        """
        return self._stat.st_size

    @property
    def lmt(self):
        # type: (LocalPath) -> int
        """mtime of file
        :param LocalPath self: this
        :rtype: int
        :return: mtime of file
        """
        return self._stat.st_mtime

    @property
    def mode(self):
        # type: (LocalPath) -> str
        """Octal file mode
        :param LocalPath self: this
        :rtype: str
        :return: octal file mode
        """
        return str(oct(self._stat.st_mode))

    @property
    def uid(self):
        # type: (LocalPath) -> int
        """Uid of file
        :param LocalPath self: this
        :rtype: int
        :return: uid of file
        """
        return self._stat.st_uid

    @property
    def gid(self):
        # type: (LocalPath) -> int
        """Gid of file
        :param LocalPath self: this
        :rtype: int
        :return: gid of file
        """
        return self._stat.st_gid


class LocalSourcePath(blobxfer.models._BaseSourcePaths):
    """Local Source Path"""

    def can_rename(self):
        # type: (LocalSourcePaths) -> bool
        """Check if source can be renamed
        :param LocalSourcePath self: this
        :rtype: bool
        :return: if rename possible
        """
        return len(self._paths) == 1 and self._paths[0].is_file()

    @staticmethod
    def is_stdin(path):
        # type: (str) -> bool
        """Check if path is stdin
        :param str path: path to check
        :rtype: bool
        :return: if path is stdin
        """
        if path == '-' or path == '/dev/stdin':
            return True
        return False

    def files(self):
        # type: (LocalSourcePaths) -> LocalPath
        """Generator for files in paths
        :param LocalSourcePath self: this
        :rtype: LocalPath
        :return: LocalPath
        """
        for _path in self._paths:
            _ppath = os.path.expandvars(os.path.expanduser(str(_path)))
            # check of path is stdin
            if blobxfer.models.upload.LocalSourcePath.is_stdin(_ppath):
                yield LocalPath(
                    parent_path=pathlib.Path(),
                    relative_path=pathlib.Path('stdin'),
                    use_stdin=True,
                )
                continue
            # resolve path
            _expath = pathlib.Path(_ppath).resolve()
            # check if path is a single file
            tmp = pathlib.Path(_ppath)
            if tmp.is_file():
                if self._inclusion_check(tmp.name):
                    yield LocalPath(
                        parent_path=tmp.parent,
                        relative_path=pathlib.Path(tmp.name),
                        use_stdin=False,
                    )
            else:
                del tmp
                for entry in blobxfer.util.scantree(_ppath):
                    _rpath = pathlib.Path(entry.path).relative_to(_ppath)
                    if not self._inclusion_check(_rpath):
                        continue
                    yield LocalPath(
                        parent_path=_expath,
                        relative_path=_rpath,
                        use_stdin=False,
                    )


class Specification(object):
    """Upload Specification"""
    def __init__(
            self, upload_options, skip_on_options, local_source_path):
        # type: (Specification, blobxfer.models.options.Upload,
        #        blobxfer.models.options.SkipOn, LocalSourcePath) -> None
        """Ctor for Specification
        :param UploadSpecification self: this
        :param blobxfer.models.options.Upload upload_options: upload options
        :param blobxfer.models.options.SkipOn skip_on_options: skip on options
        :param LocalSourcePath local_source_path: local source path
        """
        self.options = upload_options
        self.skip_on = skip_on_options
        self.destinations = []
        self.sources = local_source_path
        # validate options
        if self.options.rename:
            # ensure only one internal path is present
            if len(self.sources.paths) > 1:
                raise ValueError(
                    'cannot add more than one internal source path if rename '
                    'is specified')
            # check if internal source path is directory and rename is enabled
            if self.sources.paths[0].is_dir():
                raise ValueError(
                    'cannot rename a directory of files to upload')
        if self.options.chunk_size_bytes < 0:
            raise ValueError('chunk size cannot be negative')
        if self.options.chunk_size_bytes > _MAX_BLOCK_BLOB_CHUNKSIZE_BYTES:
            raise ValueError(
                ('chunk size value of {} exceeds maximum allowable '
                 'of {}').format(
                     self.options.chunk_size_bytes,
                     _MAX_BLOCK_BLOB_CHUNKSIZE_BYTES))
        if self.options.one_shot_bytes < 0:
            raise ValueError('one shot bytes value must be at least 0')
        if self.options.one_shot_bytes > _MAX_BLOCK_BLOB_ONESHOT_BYTES:
            raise ValueError(
                ('one shot bytes value of {} exceeds maximum allowable '
                 'of {}').format(
                     self.options.chunk_size_bytes,
                     _MAX_BLOCK_BLOB_ONESHOT_BYTES))

    def add_azure_destination_path(self, dest):
        # type: (Specification,
        #        blobxfer.operations.azure.DestinationPath) -> None
        """Add a remote Azure Destination path
        :param UploadSpecification self: this
        :param blobxfer.operations.azure.DestinationPath dest:
            Remote destination path
        """
        self.destinations.append(dest)


class Descriptor(object):
    """Upload Descriptor"""

    _AES_BLOCKSIZE = blobxfer.models.crypto.AES256_BLOCKSIZE_BYTES

    def __init__(self, lpath, ase, uid, options, resume_mgr):
        # type: (Descriptior, LocalPath,
        #        blobxfer.models.azure.StorageEntity, str,
        #        blobxfer.models.options.Upload,
        #        blobxfer.operations.resume.UploadResumeManager) -> None
        """Ctor for Descriptor
        :param Descriptor self: this
        :param LocalPath lpath: local path
        :param blobxfer.models.azure.StorageEntity ase: Azure Storage Entity
        :param str uid: unique id
        :param blobxfer.models.options.Upload options: download options
        :param blobxfer.operations.resume.UploadResumeManager resume_mgr:
            upload resume manager
        """
        self.local_path = lpath
        self.unique_id = uid
        self._offset = 0
        self._chunk_num = 0
        self._next_integrity_chunk = 0
        self._finalized = False
        self._meta_lock = threading.Lock()
        self._hasher_lock = threading.Lock()
        self._resume_mgr = resume_mgr
        self._ase = ase
        self._store_file_attr = options.store_file_properties.attributes
        self.current_iv = None
        self._initialize_encryption(options)
        # calculate the total number of ops required for transfer
        self._compute_remote_size()
        self._adjust_chunk_size(options)
        self._total_chunks = self._compute_total_chunks(self._chunk_size)
        self._outstanding_ops = self._total_chunks
        if blobxfer.util.is_not_empty(self._ase.replica_targets):
            self._outstanding_ops *= len(self._ase.replica_targets) + 1
        if self._resume_mgr:
            self._completed_chunks = bitstring.BitArray(
                length=self._total_chunks)
            self._md5_cache = {}
            self._replica_counters = {}
        # initialize integrity checkers
        self.hmac = None
        self.md5 = None
        self._initialize_integrity_checkers(options)

    @property
    def entity(self):
        # type: (Descriptor) -> blobxfer.models.azure.StorageEntity
        """Get linked blobxfer.models.azure.StorageEntity
        :param Descriptor self: this
        :rtype: blobxfer.models.azure.StorageEntity
        :return: blobxfer.models.azure.StorageEntity
        """
        return self._ase

    @property
    def must_compute_md5(self):
        # type: (Descriptor) -> bool
        """Check if MD5 must be computed
        :param Descriptor self: this
        :rtype: bool
        :return: if MD5 must be computed
        """
        return self.md5 is not None

    @property
    def all_operations_completed(self):
        # type: (Descriptor) -> bool
        """All operations are completed
        :param Descriptor self: this
        :rtype: bool
        :return: if all operations completed
        """
        with self._meta_lock:
            return self._outstanding_ops == 0

    @property
    def last_block_num(self):
        # type: (Descriptor) -> bool
        """Last used block number for block id, should only be called for
        finalize operation
        :param Descriptor self: this
        :rtype: int
        :return: block number
        """
        with self._meta_lock:
            return self._chunk_num - 1

    @property
    def is_resumable(self):
        # type: (Descriptor) -> bool
        """Upload is resume capable
        :param Descriptor self: this
        :rtype: bool
        :return: if resumable
        """
        return (self._resume_mgr is not None and self.hmac is None and
                not self.remote_is_append_blob)

    @property
    def remote_is_file(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure File
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure File
        """
        return self.entity.mode == blobxfer.models.azure.StorageModes.File

    @property
    def remote_is_page_blob(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure Page Blob
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure Page Blob
        """
        return self.entity.mode == blobxfer.models.azure.StorageModes.Page

    @property
    def remote_is_append_blob(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure Append Blob
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure Append Blob
        """
        return self.entity.mode == blobxfer.models.azure.StorageModes.Append

    @property
    def is_one_shot_block_blob(self):
        # type: (Descriptor) -> bool
        """Is one shot block blob
        :param Descriptor self: this
        :rtype: bool
        :return: if upload is a one-shot block blob
        """
        return (self._ase.mode == blobxfer.models.azure.StorageModes.Block and
                self._total_chunks == 1)

    @property
    def requires_put_block_list(self):
        # type: (Descriptor) -> bool
        """Requires a put block list operation to finalize
        :param Descriptor self: this
        :rtype: bool
        :return: if finalize requires a put block list
        """
        return (self._ase.mode == blobxfer.models.azure.StorageModes.Block and
                self._total_chunks > 1)

    @property
    def requires_non_encrypted_md5_put(self):
        # type: (Descriptor) -> bool
        """Requires a set file properties for md5 to finalize
        :param Descriptor self: this
        :rtype: bool
        :return: if finalize requires a put file properties
        """
        return (not self.entity.is_encrypted and self.must_compute_md5 and
                not self.remote_is_append_blob)

    @property
    def requires_set_file_properties_md5(self):
        # type: (Descriptor) -> bool
        """Requires a set file properties for md5 to finalize
        :param Descriptor self: this
        :rtype: bool
        :return: if finalize requires a put file properties
        """
        return (not self.entity.is_encrypted and self.must_compute_md5 and
                self.remote_is_file)

    def complete_offset_upload(self, chunk_num):
        # type: (Descriptor, int) -> None
        """Complete the upload for the offset
        :param Descriptor self: this
        :param int chunk_num: chunk num completed
        """
        with self._meta_lock:
            self._outstanding_ops -= 1
            # save resume state
            if self.is_resumable:
                # only set resumable completed if all replicas for this
                # chunk are complete
                if blobxfer.util.is_not_empty(self._ase.replica_targets):
                    if chunk_num not in self._replica_counters:
                        # start counter at -1 since we need 1 "extra" for the
                        # primary in addition to the replica targets
                        self._replica_counters[chunk_num] = -1
                    self._replica_counters[chunk_num] += 1
                    if (self._replica_counters[chunk_num] !=
                            len(self._ase.replica_targets)):
                        return
                    else:
                        self._replica_counters.pop(chunk_num)
                self._completed_chunks.set(True, chunk_num)
                completed = self._outstanding_ops == 0
                if not completed and self.must_compute_md5:
                    last_consecutive = (
                        self._completed_chunks.find('0b0')[0] - 1
                    )
                    md5digest = self._md5_cache[last_consecutive]
                else:
                    md5digest = None
                self._resume_mgr.add_or_update_record(
                    self.local_path.absolute_path, self._ase, self._chunk_size,
                    self._total_chunks, self._completed_chunks.int, completed,
                    md5digest,
                )
                # prune md5 cache
                if completed:
                    self._md5_cache.clear()
                elif (last_consecutive is not None and
                      len(self._md5_cache) >
                      _MD5_CACHE_RESUME_ENTRIES_GC_THRESHOLD):
                    mkeys = sorted(list(self._md5_cache.keys()))
                    for key in mkeys:
                        if key >= last_consecutive:
                            break
                        self._md5_cache.pop(key)

    def hmac_data(self, data):
        # type: (Descriptor, bytes) -> None
        """Send data through hmac hasher
        :param Descriptor self: this
        :param bytes data: data
        """
        with self._hasher_lock:
            self.hmac.update(data)

    def _initialize_encryption(self, options):
        # type: (Descriptor, blobxfer.models.options.Upload) -> None
        """Download is resume capable
        :param Descriptor self: this
        :param blobxfer.models.options.Upload options: upload options
        """
        if (options.rsa_public_key is not None and self.local_path.size > 0 and
                (self._ase.mode == blobxfer.models.azure.StorageModes.Block or
                 self._ase.mode == blobxfer.models.azure.StorageModes.File)):
            em = blobxfer.models.crypto.EncryptionMetadata()
            em.create_new_metadata(options.rsa_public_key)
            self.current_iv = em.content_encryption_iv
            self._ase.encryption_metadata = em

    def _compute_remote_size(self):
        # type: (Descriptor, int) -> None
        """Compute total remote file size
        :param Descriptor self: this
        :rtype: int
        :return: remote file size
        """
        size = self.local_path.size
        if size > 0:
            if self._ase.is_encrypted:
                # cipher_len_without_iv = (clear_len / aes_bs + 1) * aes_bs
                allocatesize = (size // self._AES_BLOCKSIZE + 1) * \
                    self._AES_BLOCKSIZE
            else:
                allocatesize = size
        else:
            allocatesize = 0
        self._ase.size = allocatesize
        if blobxfer.util.is_not_empty(self._ase.replica_targets):
            for rt in self._ase.replica_targets:
                rt.size = allocatesize
        logger.debug('remote size for {} is {} bytes'.format(
            self._ase.path, self._ase.size))

    def _adjust_chunk_size(self, options):
        # type: (Descriptor, blobxfer.models.options.Upload) -> None
        """Adjust chunk size for entity mode
        :param Descriptor self: this
        :param blobxfer.models.options.Upload options: upload options
        """
        chunk_size = options.chunk_size_bytes
        # auto-select chunk size
        if chunk_size == 0:
            if self._ase.mode != blobxfer.models.azure.StorageModes.Block:
                chunk_size = _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES
            else:
                if self._ase.size == 0:
                    chunk_size = _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES
                else:
                    chunk_size = _DEFAULT_AUTO_CHUNKSIZE_BYTES
                    while chunk_size < _MAX_BLOCK_BLOB_CHUNKSIZE_BYTES:
                        chunks = int(math.ceil(self._ase.size / chunk_size))
                        if chunks <= _MAX_NUM_CHUNKS:
                            break
                        chunk_size = chunk_size << 1
            logger.debug(
                'auto-selected chunk size of {} for {}'.format(
                    chunk_size, self.local_path.absolute_path))
        if self.local_path.use_stdin:
            self._chunk_size = max(
                (chunk_size, _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES)
            )
        else:
            self._chunk_size = min((chunk_size, self._ase.size))
        # ensure chunk sizes are compatible with mode
        if self._ase.mode == blobxfer.models.azure.StorageModes.Append:
            if self._chunk_size > _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES:
                self._chunk_size = _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES
                logger.debug(
                    ('adjusting chunk size to {} for append blob '
                     'from {}').format(
                         self._chunk_size, self.local_path.absolute_path))
        elif self._ase.mode == blobxfer.models.azure.StorageModes.Block:
            if (not self.local_path.use_stdin and
                    self._ase.size <= options.one_shot_bytes):
                self._chunk_size = min(
                    (self._ase.size, options.one_shot_bytes)
                )
            else:
                if self._chunk_size > _MAX_BLOCK_BLOB_CHUNKSIZE_BYTES:
                    self._chunk_size = _MAX_BLOCK_BLOB_CHUNKSIZE_BYTES
                    logger.debug(
                        ('adjusting chunk size to {} for block blob '
                         'from {}').format(
                            self._chunk_size, self.local_path.absolute_path))
        elif self._ase.mode == blobxfer.models.azure.StorageModes.File:
            if self._chunk_size > _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES:
                self._chunk_size = _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES
                logger.debug(
                    'adjusting chunk size to {} for file from {}'.format(
                        self._chunk_size, self.local_path.absolute_path))
        elif self._ase.mode == blobxfer.models.azure.StorageModes.Page:
            if self._ase.size > _MAX_PAGE_BLOB_SIZE:
                raise RuntimeError(
                    '{} size {} exceeds maximum page blob size of {}'.format(
                        self.local_path.absolute_path, self._ase.size,
                        _MAX_PAGE_BLOB_SIZE))
            if self._chunk_size > _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES:
                self._chunk_size = _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES
                logger.debug(
                    'adjusting chunk size to {} for page blob from {}'.format(
                        self._chunk_size, self.local_path.absolute_path))

    def _compute_total_chunks(self, chunk_size):
        # type: (Descriptor, int) -> int
        """Compute total number of chunks for entity
        :param Descriptor self: this
        :param int chunk_size: chunk size
        :rtype: int
        :return: num chunks
        """
        try:
            chunks = int(math.ceil(self._ase.size / chunk_size))
        except ZeroDivisionError:
            chunks = 1
        if self.local_path.use_stdin and chunks == 0:
            chunks = 1
        if (self._ase.mode != blobxfer.models.azure.StorageModes.Page and
                chunks > 50000):
            max_vector = False
            if self._ase.mode == blobxfer.models.azure.StorageModes.Block:
                if self._chunk_size == _MAX_BLOCK_BLOB_CHUNKSIZE_BYTES:
                    max_vector = True
            elif self._chunk_size == _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES:
                max_vector = True
            if max_vector:
                raise RuntimeError(
                    ('number of chunks {} exceeds maximum permissible '
                     'limit and chunk size is set at the maximum value '
                     'for {}. Please try using stripe mode '
                     'vectorization to overcome this limitation').format(
                        chunks, self.local_path.absolute_path))
            else:
                raise RuntimeError(
                    ('number of chunks {} exceeds maximum permissible '
                     'limit for {}, please adjust chunk size higher or '
                     'set to -1 for automatic chunk size selection').format(
                         chunks, self.local_path.absolute_path))
        return chunks

    def _initialize_integrity_checkers(self, options):
        # type: (Descriptor, blobxfer.models.options.Upload) -> None
        """Initialize file integrity checkers
        :param Descriptor self: this
        :param blobxfer.models.options.Upload options: upload options
        """
        if self._ase.is_encrypted:
            # ensure symmetric key exists
            if blobxfer.util.is_none_or_empty(
                    self._ase.encryption_metadata.symmetric_key):
                raise RuntimeError(
                    ('symmetric key is invalid: provide RSA private key '
                     'or metadata corrupt for {}').format(
                         self.local_path.absolute_path))
            self.hmac = self._ase.encryption_metadata.initialize_hmac()
        # both hmac and md5 can be enabled
        if (options.store_file_properties.md5 and
                not self.remote_is_append_blob):
            self.md5 = blobxfer.util.new_md5_hasher()

    def _resume(self):
        # type: (Descriptor) -> int
        """Resume upload
        :param Descriptor self: this
        :rtype: int
        :return: resume bytes
        """
        if self._resume_mgr is None or self._offset > 0:
            return None
        # check if path exists in resume db
        rr = self._resume_mgr.get_record(self._ase)
        if rr is None:
            logger.debug('no resume record for {}'.format(self._ase.path))
            return None
        # ensure lengths are the same
        if rr.length != self._ase.size:
            logger.warning('resume length mismatch {} -> {}'.format(
                rr.length, self._ase.size))
            return None
        # compute replica factor
        if blobxfer.util.is_not_empty(self._ase.replica_targets):
            replica_factor = 1 + len(self._ase.replica_targets)
        else:
            replica_factor = 1
        # set offsets if completed
        if rr.completed:
            with self._meta_lock:
                logger.debug('{} upload already completed'.format(
                    self._ase.path))
                self._offset = rr.total_chunks * rr.chunk_size
                self._chunk_num = rr.total_chunks
                self._chunk_size = rr.chunk_size
                self._total_chunks = rr.total_chunks
                self._completed_chunks.int = rr.completed_chunks
                self._outstanding_ops = 0
                return self._src_ase.size * replica_factor
        # encrypted files are not resumable due to hmac requirement
        if self._ase.is_encrypted:
            logger.debug('cannot resume encrypted entity {}'.format(
                self._ase.path))
            return None
        # check if path exists
        if not pathlib.Path(rr.local_path).exists():
            logger.warning('resume from local path {} does not exist'.format(
                rr.local_path))
            return None
        # re-hash from 0 to offset if needed
        _cc = bitstring.BitArray(length=rr.total_chunks)
        _cc.int = rr.completed_chunks
        curr_chunk = _cc.find('0b0')[0]
        del _cc
        _fd_offset = 0
        _end_offset = min((curr_chunk * rr.chunk_size, rr.length))
        if self.md5 is not None and curr_chunk > 0:
            _blocksize = blobxfer.util.MEGABYTE << 2
            logger.debug(
                'integrity checking existing file {} offset {} -> {}'.format(
                    self._ase.path,
                    self.local_path.view.fd_start,
                    self.local_path.view.fd_start + _end_offset)
            )
            with self._hasher_lock:
                with self.local_path.absolute_path.open('rb') as filedesc:
                    filedesc.seek(self.local_path.view.fd_start, 0)
                    while _fd_offset < _end_offset:
                        if (_fd_offset + _blocksize) > _end_offset:
                            _blocksize = _end_offset - _fd_offset
                        _buf = filedesc.read(_blocksize)
                        self.md5.update(_buf)
                        _fd_offset += _blocksize
            del _blocksize
            # compare hashes
            hexdigest = self.md5.hexdigest()
            if rr.md5hexdigest != hexdigest:
                logger.warning(
                    'MD5 mismatch resume={} computed={} for {}'.format(
                        rr.md5hexdigest, hexdigest, self._ase.path))
                # reset hasher
                self.md5 = blobxfer.util.new_md5_hasher()
                return None
        # set values from resume
        with self._meta_lock:
            self._offset = _end_offset
            self._chunk_num = curr_chunk
            self._chunk_size = rr.chunk_size
            self._total_chunks = rr.total_chunks
            self._completed_chunks = bitstring.BitArray(length=rr.total_chunks)
            self._completed_chunks.set(True, range(0, curr_chunk + 1))
            self._outstanding_ops = (
                (rr.total_chunks - curr_chunk) * replica_factor
            )
            logger.debug(
                ('resuming file {} from byte={} chunk={} chunk_size={} '
                 'total_chunks={} outstanding_ops={}').format(
                     self._ase.path, self._offset, self._chunk_num,
                     self._chunk_size, self._total_chunks,
                     self._outstanding_ops))
            return _end_offset * replica_factor

    def next_offsets(self):
        # type: (Descriptor) -> Offsets
        """Retrieve the next offsets
        :param Descriptor self: this
        :rtype: Offsets
        :return: upload offsets
        """
        resume_bytes = self._resume()
        with self._meta_lock:
            if self._chunk_num >= self._total_chunks:
                return None, resume_bytes
            if self._offset + self._chunk_size > self._ase.size:
                num_bytes = self._ase.size - self._offset
            else:
                num_bytes = self._chunk_size
            chunk_num = self._chunk_num
            range_start = self._offset
            range_end = self._offset + num_bytes - 1
            self._offset += num_bytes
            self._chunk_num += 1
            if self._ase.is_encrypted and self._offset >= self._ase.size:
                pad = True
            else:
                pad = False
            return Offsets(
                chunk_num=chunk_num,
                num_bytes=num_bytes,
                range_start=range_start,
                range_end=range_end,
                pad=pad,
            ), resume_bytes

    def read_data(self, offsets):
        # type: (Descriptor, Offsets) -> Tuple[bytes, Offsets]
        """Read data from file
        :param Descriptor self: this
        :param Offsets offsets: offsets
        :rtype: tuple
        :return: (file data bytes, new Offsets if stdin)
        """
        newoffset = None
        if not self.local_path.use_stdin:
            if offsets.num_bytes == 0:
                return None, None
            # compute start from view
            start = self.local_path.view.fd_start + offsets.range_start
            # encrypted offsets will read past the end of the file due
            # to padding, but will be accounted for after encryption+padding
            with self.local_path.absolute_path.open('rb') as fd:
                fd.seek(start, 0)
                data = fd.read(offsets.num_bytes)
        else:
            data = blobxfer.STDIN.read(self._chunk_size)
            if not data:
                with self._meta_lock:
                    self._total_chunks -= 1
                    self._chunk_num -= 1
                    self._outstanding_ops -= 1
            else:
                num_bytes = len(data)
                with self._meta_lock:
                    newoffset = Offsets(
                        chunk_num=self._chunk_num - 1,
                        num_bytes=num_bytes,
                        range_start=self._offset,
                        range_end=self._offset + num_bytes - 1,
                        pad=False,
                    )
                    self._total_chunks += 1
                    self._outstanding_ops += 1
                    self._offset += num_bytes
                    self._ase.size += num_bytes
        if self.must_compute_md5 and data:
            with self._hasher_lock:
                self.md5.update(data)
                if self.is_resumable:
                    self._md5_cache[self._chunk_num - 1] = self.md5.hexdigest()
        return data, newoffset

    def generate_metadata(self):
        # type: (Descriptor) -> dict
        """Generate metadata for descriptor
        :param Descriptor self: this
        :rtype: dict or None
        :return: kv metadata dict
        """
        genmeta = {}
        encmeta = {}
        # generate encryption metadata
        if self._ase.is_encrypted:
            if self.must_compute_md5:
                md5digest = blobxfer.util.base64_encode_as_string(
                    self.md5.digest())
            else:
                md5digest = None
            if self.hmac is not None:
                hmacdigest = blobxfer.util.base64_encode_as_string(
                    self.hmac.digest())
            else:
                hmacdigest = None
            encmeta = self._ase.encryption_metadata.convert_to_json_with_mac(
                md5digest, hmacdigest)
        # generate file attribute metadata
        if self._store_file_attr and not self.local_path.use_stdin:
            merged = blobxfer.models.metadata.generate_fileattr_metadata(
                self.local_path, genmeta)
            if merged is not None:
                genmeta = merged
        # generate vectored io metadata
        if self.local_path.view.mode == VectoredIoDistributionMode.Stripe:
            merged = blobxfer.models.metadata.\
                generate_vectored_io_stripe_metadata(self.local_path, genmeta)
            if merged is not None:
                genmeta = merged
        if len(encmeta) > 0:
            metadata = encmeta
        else:
            metadata = {}
        if len(genmeta) > 0:
            metadata[blobxfer.models.metadata.JSON_KEY_BLOBXFER_METADATA] = \
                json.dumps(genmeta, ensure_ascii=False, sort_keys=True)
        if len(metadata) == 0:
            return None
        return metadata



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
import logging
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
# non-stdlib imports
import azure.common
import azure.storage.file
# local imports
import blobxfer.retry
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)


def create_client(storage_account, timeout):
    # type: (blobxfer.operations.azure.StorageAccount, tuple) -> FileService
    """Create file client
    :param blobxfer.operations.azure.StorageAccount storage_account:
        storage account
    :param tuple timeout: timeout tuple
    :rtype: FileService
    :return: file service client
    """
    if storage_account.is_sas:
        client = azure.storage.file.FileService(
            account_name=storage_account.name,
            sas_token=storage_account.key,
            endpoint_suffix=storage_account.endpoint,
            request_session=storage_account.session,
            socket_timeout=timeout)
    else:
        client = azure.storage.file.FileService(
            account_name=storage_account.name,
            account_key=storage_account.key,
            endpoint_suffix=storage_account.endpoint,
            request_session=storage_account.session,
            socket_timeout=timeout)
    # set retry policy
    client.retry = blobxfer.retry.ExponentialRetryWithMaxWait().retry
    return client


def parse_file_path(filepath):
    # type: (pathlib.Path) -> Tuple[str, str]
    """Parse file path from file path
    :param str filepath: file path
    :rtype: tuple
    :return: (dirname, rest of path)
    """
    if not isinstance(filepath, pathlib.Path):
        filepath = pathlib.Path(filepath)
    dirname = '/'.join(filepath.parts[:len(filepath.parts) - 1])
    if len(dirname) == 0:
        dirname = None
    if len(filepath.parts) > 0:
        fname = filepath.parts[-1]
    else:
        fname = None
    return (dirname, fname)


def get_file_properties(client, fileshare, prefix, timeout=None):
    # type: (azure.storage.file.FileService, str, str, int) ->
    #        azure.storage.file.models.File
    """Get file properties
    :param FileService client: blob client
    :param str fileshare: file share name
    :param str prefix: path prefix
    :param int timeout: timeout
    :rtype: azure.storage.file.models.File
    :return: file properties
    """
    dirname, fname = parse_file_path(prefix)
    try:
        return client.get_file_properties(
            share_name=fileshare,
            directory_name=dirname,
            file_name=fname,
            timeout=timeout,
        )
    except azure.common.AzureMissingResourceHttpError:
        return None


def check_if_single_file(client, fileshare, prefix, timeout=None):
    # type: (azure.storage.file.FileService, str, str, int) ->
    #        Tuple[bool, azure.storage.file.models.File]
    """Check if prefix is a single file or multiple files
    :param FileService client: blob client
    :param str fileshare: file share name
    :param str prefix: path prefix
    :param int timeout: timeout
    :rtype: tuple
    :return: (if prefix in fileshare is a single file, file)
    """
    if blobxfer.util.is_none_or_empty(prefix):
        return (False, None)
    file = get_file_properties(client, fileshare, prefix, timeout)
    if file is None:
        return (False, file)
    else:
        return (True, file)


def list_files(client, fileshare, prefix, recursive, timeout=None):
    # type: (azure.storage.file.FileService, str, str, bool, int) ->
    #        azure.storage.file.models.File
    """List files in path
    :param azure.storage.file.FileService client: file client
    :param str fileshare: file share
    :param str prefix: path prefix
    :param bool recursive: recursive
    :param int timeout: timeout
    :rtype: azure.storage.file.models.File
    :return: generator of files
    """
    # if single file, then yield file and return
    _check = check_if_single_file(client, fileshare, prefix, timeout)
    if _check[0]:
        yield _check[1]
        return
    # else recursively list from prefix path
    dirs = [prefix]
    while len(dirs) > 0:
        dir = dirs.pop()
        files = client.list_directories_and_files(
            share_name=fileshare,
            directory_name=dir,
            timeout=timeout,
        )
        for file in files:
            fspath = str(
                pathlib.Path(dir if dir is not None else '') / file.name)
            if type(file) == azure.storage.file.models.File:
                fsprop = client.get_file_properties(
                    share_name=fileshare,
                    directory_name=None,
                    file_name=fspath,
                    timeout=timeout,
                )
                yield fsprop
            else:
                if recursive:
                    dirs.append(fspath)


def list_all_files(client, fileshare, timeout=None):
    # type: (azure.storage.file.FileService, str, int) -> str
    """List all files in share
    :param azure.storage.file.FileService client: file client
    :param str fileshare: file share
    :param int timeout: timeout
    :rtype: str
    :return: file name
    """
    dirs = [None]
    while len(dirs) > 0:
        dir = dirs.pop()
        files = client.list_directories_and_files(
            share_name=fileshare,
            directory_name=dir,
            timeout=timeout,
        )
        for file in files:
            fspath = str(
                pathlib.Path(dir if dir is not None else '') / file.name)
            if type(file) == azure.storage.file.models.File:
                yield fspath
            else:
                dirs.append(fspath)


def delete_file(client, fileshare, name, timeout=None):
    # type: (azure.storage.file.FileService, str, str, int) -> None
    """Delete file from share
    :param azure.storage.file.FileService client: file client
    :param str fileshare: file share
    :param str name: file name
    :param int timeout: timeout
    """
    dir, fpath = parse_file_path(name)
    client.delete_file(
        share_name=fileshare,
        directory_name=dir,
        file_name=fpath,
        timeout=timeout,
    )


def get_file_range(ase, offsets, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity,
    #        blobxfer.models.download.Offsets, int) -> bytes
    """Retrieve file range
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param blobxfer.models.download.Offsets offsets: download offsets
    :param int timeout: timeout
    :rtype: bytes
    :return: content for file range
    """
    dir, fpath = parse_file_path(ase.name)
    return ase.client._get_file(
        share_name=ase.container,
        directory_name=dir,
        file_name=fpath,
        start_range=offsets.range_start,
        end_range=offsets.range_end,
        validate_content=False,  # HTTPS takes care of integrity during xfer
        timeout=timeout,
    ).content


def create_share(ase, containers_created, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, dict, int) -> None
    """Create file share
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param dict containers_created: containers already created map
    :param int timeout: timeout
    """
    # check if auth allows create container
    if not ase.create_containers:
        return
    key = ase.client.account_name + ':file=' + ase.container
    if key not in containers_created:
        try:
            ase.client.create_share(
                share_name=ase.container,
                fail_on_exist=True,
                timeout=timeout)
        except azure.common.AzureConflictHttpError:
            pass
        else:
            containers_created.add(key)
            logger.info('created file share {} on storage account {}'.format(
                ase.container, ase.client.account_name))


def create_all_parent_directories(ase, dirs_created, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, dict, int) -> None
    """Create all parent directories for a file
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param dict dirs_created: directories already created map
    :param int timeout: timeout
    """
    dirs = pathlib.Path(ase.name).parts
    if len(dirs) <= 1:
        return
    # remove last part (which is the file)
    dirs = dirs[:-1]
    dk = ase.client.account_name + ':' + ase.container
    for i in range(0, len(dirs)):
        dir = str(pathlib.Path(*(dirs[0:i + 1])))
        if dk not in dirs_created or dir not in dirs_created[dk]:
            ase.client.create_directory(
                share_name=ase.container,
                directory_name=dir,
                fail_on_exist=False,
                timeout=timeout)
            if dk not in dirs_created:
                dirs_created[dk] = set()
            dirs_created[dk].add(dir)


def create_file(ase, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, int) -> None
    """Create file remotely
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param int timeout: timeout
    """
    dir, fpath = parse_file_path(ase.name)
    ase.client.create_file(
        share_name=ase.container,
        directory_name=dir,
        file_name=fpath,
        content_length=ase.size,
        content_settings=azure.storage.file.models.ContentSettings(
            content_type=blobxfer.util.get_mime_type(fpath)
        ),
        timeout=timeout)


def put_file_range(ase, offsets, data, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity,
    #        blobxfer.models.upload.Offsets, bytes, int) -> None
    """Puts a range of bytes into the remote file
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param blobxfer.models.upload.Offsets offsets: upload offsets
    :param bytes data: data
    :param int timeout: timeout
    """
    dir, fpath = parse_file_path(ase.name)
    ase.client.update_range(
        share_name=ase.container,
        directory_name=dir,
        file_name=fpath,
        data=data,
        start_range=offsets.range_start,
        end_range=offsets.range_end,
        validate_content=False,  # integrity is enforced with HTTPS
        timeout=timeout)


def set_file_md5(ase, md5, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, str, int) -> None
    """Set file properties MD5
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param str md5: md5 as base64
    :param int timeout: timeout
    """
    dir, fpath = parse_file_path(ase.name)
    ase.client.set_file_properties(
        share_name=ase.container,
        directory_name=dir,
        file_name=fpath,
        content_settings=azure.storage.file.models.ContentSettings(
            content_type=blobxfer.util.get_mime_type(fpath),
            content_md5=md5,
        ),
        timeout=timeout)


def set_file_metadata(ase, metadata, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, dict, int) -> None
    """Set file metadata
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param dict metadata: metadata kv pairs
    :param int timeout: timeout
    """
    dir, fpath = parse_file_path(ase.name)
    ase.client.set_file_metadata(
        share_name=ase.container,
        directory_name=dir,
        file_name=fpath,
        metadata=metadata,
        timeout=timeout)



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
import logging
# non-stdlib imports
import azure.storage.blob
# local imports
import blobxfer.retry

# create logger
logger = logging.getLogger(__name__)


def create_client(storage_account, timeout):
    # type: (blobxfer.operations.azure.StorageAccount,
    #        tuple) -> AppendBlobService
    """Create Append blob client
    :param blobxfer.operations.azure.StorageAccount storage_account:
        storage account
    :param tuple timeout: timeout tuple
    :rtype: AppendBlobService
    :return: append blob service client
    """
    if storage_account.is_sas:
        client = azure.storage.blob.AppendBlobService(
            account_name=storage_account.name,
            sas_token=storage_account.key,
            endpoint_suffix=storage_account.endpoint,
            request_session=storage_account.session,
            socket_timeout=timeout)
    else:
        client = azure.storage.blob.AppendBlobService(
            account_name=storage_account.name,
            account_key=storage_account.key,
            endpoint_suffix=storage_account.endpoint,
            request_session=storage_account.session,
            socket_timeout=timeout)
    # set retry policy
    client.retry = blobxfer.retry.ExponentialRetryWithMaxWait().retry
    return client


def create_blob(ase, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, int) -> None
    """Create append blob
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param int timeout: timeout
    """
    ase.client.create_blob(
        container_name=ase.container,
        blob_name=ase.name,
        content_settings=azure.storage.blob.models.ContentSettings(
            content_type=blobxfer.util.get_mime_type(ase.name)
        ),
        timeout=timeout)  # noqa


def append_block(ase, data, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, bytes, int) -> None
    """Appends a block into remote blob
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param bytes data: data
    :param int timeout: timeout
    """
    ase.client.append_block(
        container_name=ase.container,
        blob_name=ase.name,
        block=data,
        validate_content=False,  # integrity is enforced with HTTPS
        timeout=timeout)  # noqa



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import logging
import multiprocessing
import threading
try:
    import queue
except ImportError:  # noqa
    import Queue as queue

# create logger
logger = logging.getLogger(__name__)


class _MultiprocessOffload(object):
    def __init__(self, target, num_workers, description=None):
        # type: (_MultiprocessOffload, function, int, str) -> None
        """Ctor for Crypto Offload
        :param _MultiprocessOffload self: this
        :param function target: target function for process
        :param int num_workers: number of worker processes
        :param str description: description
        """
        self._task_queue = multiprocessing.Queue()
        self._done_queue = multiprocessing.Queue()
        self._done_cv = multiprocessing.Condition()
        self._term_signal = multiprocessing.Value('i', 0)
        self._procs = []
        self._check_thread = None
        self._initialize_processes(target, num_workers, description)

    @property
    def done_cv(self):
        # type: (_MultiprocessOffload) -> multiprocessing.Condition
        """Get Done condition variable
        :param _MultiprocessOffload self: this
        :rtype: multiprocessing.Condition
        :return: cv for download done
        """
        return self._done_cv

    @property
    def terminated(self):
        # type: (_MultiprocessOffload) -> bool
        """Check if terminated
        :param _MultiprocessOffload self: this
        :rtype: bool
        :return: if terminated
        """
        return self._term_signal.value == 1

    def _initialize_processes(self, target, num_workers, description):
        # type: (_MultiprocessOffload, function, int, str) -> None
        """Initialize processes
        :param _MultiprocessOffload self: this
        :param function target: target function for process
        :param int num_workers: number of worker processes
        :param str description: description
        """
        if num_workers is None or num_workers < 1:
            raise ValueError('invalid num_workers: {}'.format(num_workers))
        logger.debug('initializing {}{} processes'.format(
            num_workers, ' ' + description if not None else ''))
        for _ in range(num_workers):
            proc = multiprocessing.Process(target=target)
            proc.start()
            self._procs.append(proc)

    def finalize_processes(self):
        # type: (_MultiprocessOffload) -> None
        """Finalize processes
        :param _MultiprocessOffload self: this
        """
        self._term_signal.value = 1
        if self._check_thread is not None:
            self._check_thread.join()
        for proc in self._procs:
            proc.join()

    def pop_done_queue(self):
        # type: (_MultiprocessOffload) -> object
        """Get item from done queue
        :param _MultiprocessOffload self: this
        :rtype: object or None
        :return: object from done queue, if exists
        """
        try:
            return self._done_queue.get_nowait()
        except queue.Empty:
            return None

    def initialize_check_thread(self, check_func):
        # type: (_MultiprocessOffload, function) -> None
        """Initialize the crypto done queue check thread
        :param Downloader self: this
        :param function check_func: check function
        """
        self._check_thread = threading.Thread(target=check_func)
        self._check_thread.start()



from codecs import open
import os
import re
try:
    from setuptools import setup
except ImportError:  # noqa
    from distutils.core import setup
import sys

if sys.argv[-1] == 'publish':
    os.system('rm -rf blobxfer.egg-info/ build dist __pycache__/')
    os.system('python setup.py sdist bdist_wheel')
    sys.exit()
elif sys.argv[-1] == 'upload':
    os.system('twine upload dist/*')
    sys.exit()
elif sys.argv[-1] == 'sdist' or sys.argv[-1] == 'bdist_wheel':
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
else:
    long_description = ''

with open('blobxfer/version.py', 'r', 'utf-8') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(), re.MULTILINE).group(1)

if not version or len(version) == 0:
    raise RuntimeError('Cannot find version')

packages = [
    'blobxfer',
    'blobxfer.models',
    'blobxfer.operations',
    'blobxfer.operations.azure',
    'blobxfer.operations.azure.blob',
    'blobxfer_cli',
]

install_requires = [
    'azure-storage==0.36.0',
    'bitstring==3.1.5',
    'click==6.7',
    'cryptography>=2.0.3',
    'future==0.16.0',
    'pathlib2==2.3.0;python_version<"3.5"',
    'python-dateutil==2.6.1',
    'requests==2.18.4',
    'ruamel.yaml==0.15.32',
    'scandir==1.5;python_version<"3.5"',
]

setup(
    name='blobxfer',
    version=version,
    author='Microsoft Corporation, Azure Batch and HPC Team',
    author_email='',
    description='Azure storage transfer tool and library',
    long_description=long_description,
    platforms='any',
    url='https://github.com/Azure/blobxfer',
    license='MIT',
    packages=packages,
    package_data={'blobxfer': ['LICENSE']},
    package_dir={'blobxfer': 'blobxfer', 'blobxfer_cli': 'cli'},
    entry_points={
        'console_scripts': 'blobxfer=blobxfer_cli.cli:cli',
    },
    zip_safe=False,
    install_requires=install_requires,
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],
    keywords='azcopy azure storage blob files transfer copy smb cifs',
)
